# -*- coding: utf-8 -*-

import numpy as np
import random as r


__all__ = ['Interval','I','Logical']

class Interval:
    """
    Interval
    ---------
    An interval is an uncertain number for which only the endpoints are known, for example if :math:`x=[a,b]`
    then this can be interpreted as :math:`x` being between :math:`a` and :math:`b` but with no more information about the value of :math:`x`
    .
    Intervals can be created using::
        pba.I(left,right)
        pba.Interval(left,right)
    Parameters
    ----------
    left : numeric
        left side of interval
    right : numeric
        right side of interval
    Attributes
    ----------
    left : numeric
        left side of interval
    right : numeric
        right side of interval

    """
    def __init__(self,left = None, right = None):


        # kill complex nums
        assert not isinstance(left, np.complex) or not isinstance(right, np.complex), "Inputs must be real numbers"

        # assume vaccous if no inputs
        if left is None and right is None:
            right = np.inf
            left = -np.inf

        # If only one input assume zero width
        elif left is None and right is not None:
            left = right
        elif left is not None and right is None:
            right = left

        # if iterable, find endpoints
        if hasattr(left, '__iter__') and hasattr(right, '__iter__'):

            LL = min(left)
            UL = min(right)
            LU = max(left)
            UU = max(right)

            left = min(LL,LU)
            right = max(LU,UU)

        elif hasattr(left, '__iter__'):

            LL = min(left)
            LU = max(left)

            left = min(LL,LU)


        elif hasattr(right, '__iter__'):

            UL = min(right)
            UU = max(right)

            right = max(UL,UU)


        if left > right:
            LowerUpper = [left, right]
            left = min(LowerUpper)
            right = max(LowerUpper)

        self.left = left
        self.right = right

    def __repr__(self) -> str: # return
        return "Interval [%g, %g]"%(self.left,self.right)

    def __str__(self) -> str: # print
        return "Interval [%g, %g]"%(self.left,self.right)

    def __format__(self, format_spec: str) -> str:
        try:
            return f'[{format(self.left, format_spec)},{format(self.right, format_spec)}]'
        except:
            raise ValueError(f'{format_spec} format specifier not understood for Interval object')
        
    def __iter__(self):
        for bound in [self.left, self.right]:
            yield bound

    def __len__(self):
        return 2

    def __add__(self,other):

        if other.__class__.__name__ == 'Interval':
            lo = self.left + other.left
            hi = self.right + other.right
        elif other.__class__.__name__ == 'Pbox':
            # Perform Pbox addition assuming independance
            return other.add(self, method = 'i')
        else:
            try:
                lo = self.left + other
                hi  = self.right + other
            except:
                return NotImplemented

        return Interval(lo,hi)

    def __radd__(self,left):
        return self.__add__(left)

    def __sub__(self, other):

        if other.__class__.__name__ == "Interval":

            lo = self.left - other.right
            hi = self.right - other.left
        elif other.__class__.__name__ == "Pbox":
            # Perform Pbox subtractnion assuming independance
            return other.rsub(self)
        else:
            try:
                lo = self.left - other
                hi  = self.right - other
            except:
                return NotImplemented

        return Interval(lo,hi)

    def __rsub__(self, other):
        if other.__class__.__name__ == "Interval":
            # should be overkill
            lo = other.right - self.left
            hi = other.right - self.right

        elif other.__class__.__name__ == "Pbox":
            # shoud have be caught by Pbox.__sub__()
            return other.__sub__(self)
        else:
            try:
                lo = other - self.right
                hi = other - self.left

            except:
                return NotImplemented

        return Interval(lo,hi)

    def __neg__(self):
        return Interval(-self.right, -self.left)
    
    def __mul__(self,other):
        if other.__class__.__name__ == "Interval":

            b1 = self.lo() * other.lo()
            b2 = self.lo() * other.hi()
            b3 = self.hi() * other.lo()
            b4 = self.hi() * other.hi()

            lo = min(b1,b2,b3,b4)
            hi = max(b1,b2,b3,b4)

        elif other.__class__.__name__ == "Pbox":

            return other.mul(self)

        else:

            try:

                lo = self.lo() * other
                hi = self.hi() * other

            except:

                return NotImplemented

        return Interval(lo,hi)

    def __rmul__(self,other):
        return self * other

    def __truediv__(self,other):

        if other.__class__.__name__ == "Interval":

            if other.straddles_zero():
                if other.left == 0:
                    lo = min(self.lo() / other.hi(), self.hi() / other.hi())
                    hi = np.inf
                elif other.right == 0:
                    lo = -np.inf
                    hi = max(self.lo() / other.lo(), self.hi() / other.lo())
                else:
                    # Cant divide by zero
                    raise ZeroDivisionError()
            else:
                b1 = self.lo() / other.lo()
                b2 = self.lo() / other.hi()
                b3 = self.hi() / other.lo()
                b4 = self.hi() / other.hi()

                lo = min(b1,b2,b3,b4)
                hi = max(b1,b2,b3,b4)

        else:
            try:
                lo = self.lo()/other
                hi = self.hi()/other
            except:

                return NotImplemented

        return Interval(lo,hi)


    def __rtruediv__(self,other):
        
        if self.straddles_zero():
            
            raise ZeroDivisionError()
        
        try:
            return other * self.recip()
        except:
            return NotImplemented


    def __pow__(self,other):
        if other.__class__.__name__ == "Interval":
            pow1 = self.left ** other.left
            pow2 = self.left ** other.right
            pow3 = self.right ** other.left
            pow4 = self.right ** other.right
            powUp = max(pow1,pow2,pow3,pow4)
            powLow = min(pow1,pow2,pow3,pow4)
        elif other.__class__.__name__ in ("int", "float"):
            pow1 = self.left ** other
            pow2 = self.right ** other
            powUp = max(pow1,pow2)
            powLow = min(pow1,pow2)
            if (self.right >= 0) and (self.left <= 0) and (other % 2 == 0):
                powLow = 0
        return Interval(powLow,powUp)

    def __rpow__(self,left):
        if left.__class__.__name__ == "Interval":
            pow1 = left.left ** self.left
            pow2 = left.left ** self.right
            pow3 = left.right ** self.left
            pow4 = left.right ** self.right
            powUp = max(pow1,pow2,pow3,pow4)
            powLow = min(pow1,pow2,pow3,pow4)

        elif left.__class__.__name__ in ("int", "float"):
            pow1 = left ** self.left
            pow2 = left ** self.right
            powUp = max(pow1,pow2)
            powLow = min(pow1,pow2)

        return Interval(powLow,powUp)


    def __lt__(self,other):
        # <
        if other.__class__.__name__ == 'Interval':
            if self.right < other.left:
                return Logical(1,1)
            elif self.left > other.right:
                return Logical(0,0)
            elif self.straddles(other.left,endpoints = False) or self.straddles(other.right,endpoints = False):
                return Logical(0,1)
            else:
                return Logical(0,0)
        else:
            try:
                if self.right < other:
                    return Logical(1,1)
                elif self.straddles(other,endpoints = False):
                    return Logical(0,1)
                else:
                    return Logical(0,0)
            except Exception as e:
                raise ValueError

    def __eq__(self,other):
        # ==
        if other.__class__.__name__ == 'Interval':
            if self.straddles(other.left) or self.straddles(other.right):
                return Logical(0,1)
            else:
                return Logical(0,0)
        elif other is None:
            try:
                self is None
            except:
                raise ValueError
        else:
            try:
                if self.straddles(other):
                    return Logical(0,1)
                else:
                    return Logical(0,0)
            except:
                raise ValueError


    def __gt__(self,other):
        # >
        if other.__class__.__name__ == 'Interval':
            if self.right < other.left:
                return Logical(0,0)
            elif self.left > other.right:
                return Logical(1,1)
            elif self.straddles(other.left,endpoints = False) or self.straddles(other.right,endpoints = False):
                return Logical(0,1)
            else:
                return Logical(0,0)
        else:
            try:
                if self.left > other:
                    return Logical(1,1)
                elif self.straddles(other,endpoints = False):
                    return Logical(0,1)
                else:
                    return Logical(0,0)
            except Exception as e:
                raise ValueError

    def __ne__(self,other):
        # !=
        if other.__class__.__name__ == 'Interval':
            if self.straddles(other.left) or self.straddles(other.right):
                return Logical(0,1)
            else:
                return Logical(1,1)
        else:
            try:
                if self.straddles(other):
                    return Logical(0,1)
                else:
                    return Logical(1,1)
            except:
                try:
                    return not self is other
                except:
                    raise ValueError

    def __le__(self,other):
        # <=
        if other.__class__.__name__ == 'Interval':
            if self.right <= other.left:
                return Logical(1,1)
            elif self.left >= other.right:
                return Logical(0,0)
            elif self.straddles(other.left,endpoints = True) or self.straddles(other.right,endpoints = True):
                return Logical(0,1)
            else:
                return Logical(0,0)
        else:
            try:
                if self.right <= other:
                    return Logical(1,1)
                elif self.straddles(other,endpoints = True):
                    return Logical(0,1)
                else:
                    return Logical(0,0)
            except Exception as e:
                raise ValueError

    def __ge__(self,other):
        if other.__class__.__name__ == 'Interval':
            if self.right <= other.left:
                return Logical(0,0)
            elif self.left >= other.right:
                return Logical(1,1)
            elif self.straddles(other.left,endpoints = True) or self.straddles(other.right,endpoints = True):
                return Logical(0,1)
            else:
                return Logical(0,0)
        else:
            try:
                if self.left > other:
                    return Logical(1,1)
                elif self.straddles(other,endpoints = True):
                    return Logical(0,1)
                else:
                    return Logical(0,0)
            except Exception as e:
                raise ValueError

    def __bool__(self):

        try:
            if self.to_logical():
                return True
            else:
                return False
        except:
            raise ValueError("Truth value of Interval %s is ambiguous" %self)
        
    def __abs__(self):
        if self.straddles_zero():
            return Interval(0, max(abs(self.left),abs(self.right)))
        else:
            return Interval(abs(self.left),abs(self.right))

    def padd(self,other):
        """
        Returns addition using perfect arithmetic
        
        a+b = [a.left + b.left, a.right + b.right]
        """
        return Interval(self.left + other.left, self.right + other.right)
  
    def psub(self,other):
        """
        Returns subtraction using perfect arithmetic
        
        a+b = [a.left - b.left, a.right - b.right]
        """
        return Interval(self.left - other.left, self.right - other.right)

    def pmul(self,other):
        """
        Returns multiplication using opposite arithmetic
        
        a*b = [a.left * b.left, a.right * b.right]
        """
        return Interval(self.left * other.left, self.right * other.right)

    def pdiv(self,other):
        """
        Returns division using opposite arithmetic
        
        a/b = [a.left / b.left, a.right / b.right]
        """
        return Interval(self.left / other.left, self.right / other.right)    
    
    def oadd(self,other):
        """
        Returns addition using opposite arithmetic
        
        a+b = [a.left + b.right, a.right + b.left]
        """
        return Interval(self.left + other.right, self.right + other.left)
  
    def osub(self,other):
        """
        Returns subtraction using opposite arithmetic
        
        a+b = [a.left - b.right, a.right - b.left]
        """
        return Interval(self.left - other.right, self.right - other.left)
    
    def omul(self,other):
        """
        Returns multiplication using opposite arithmetic
        
        a*b = [a.left * b.right, a.right * b.left]
        """
        return Interval(self.left * other.right, self.right * other.left)

    def odiv(self,other):
        """
        Returns division using opposite arithmetic
        
        a/b = [a.left / b.right, a.right / b.left]
        """
        return Interval(self.left / other.right, self.right / other.left)      
    
    def equiv(self,other):
        """
        Checks whether two intervals are equivalent. 
        True if self.left == other.right and self.right == other.right
        """
        return (self.left == other.left and self.right == other.right)
    
    def lo(self):
        """
        Returns the left side of the interval
        """
        return self.left

    def hi(self):
        """
        Returns the right side of the interval
        """
        return self.right
    
    def width(self):
        """
        Returns the width of the interval
        self.right - self.left
        """
        return self.right - self.left

    
    def midpoint(self):
        """
        Returns midpoint of interval
        (self.left+self.right)/2
        """
        
        return (self.left+self.right)/2
    
    def to_logical(self):
        '''
        Turns the interval into a logical interval, this is done by chacking the truth value of the ends of the interval
        '''
        if self.left:
            left = True
        else:
            left = False
        
        if self.right:
            right = True
        else:
            right = False
        
        return Logical(left,right)
    
        
    def straddles(self,N, endpoints = True):
        """
        Parameters
        ----------
        N : numeric
            Number to check
        endpoints : bool
            Whether to include the endpoints within the check

        Returns
        -------
        True
            If :math:`\\mathrm{left} \\leq N \\leq \mathrm{right}` (Assuming `endpoints=True`)
        False
            Otherwise
        """
        if endpoints:
            if self.left <= N and self.right >= N:
                return True
        else:
            if self.left < N and self.right > N:
                return True

        return False

    def straddles_zero(self,endpoints = True):
        """
        Checks whether :math:`0` is within the interval
        """
        return self.straddles(0,endpoints)

    def recip(self):
        """
        Calculates the reciprocle of the interval.
        If :math:`0 \\in [a,b]` it returns a division by zero error

        """
        if self.straddles_zero():
            # Cant divide by zero
            raise ZeroDivisionError()

        elif 1/self.hi() < 1/self.lo():
            return Interval(1/self.hi(), 1/self.lo())
        else:
            return Interval(1/self.lo(), 1/self.hi())

    def intersection(self, other):
        '''
        Calculates the intersection between two intervals
        '''
        if isinstance(other, Interval):
            if self.straddles(other):
                return I(max([x.left for x in [self, other]]), min([x.right for x in [self, other]]))
            else:
                return None
        elif isinstance(other, list):
            if all([self.straddles(o) for o in other]):
                assert all([isinstance(o, Interval) for o in other]), 'All intersected objects must be intervals'
                return I(max([x.left for x in [self] + other]), min([x.right for x in [self] + other]))
            else:
                return None
        else:
            if self.straddles(other):
                return other
            else:
                return None

    def exp(self):
        lo = np.exp(self.left)
        hi = np.exp(self.right)
        return Interval(lo,hi)
# a = Interval(1,2)
# b = Interval(3,4)
# c = Interval(-2,5)
# d = Interval(-7,-4)

##.sort() function to sort numbers for median
##list1.count(x) function to help with mode

# Alias
I = Interval

class Logical(Interval):
    '''
    Imprecise Boolean object
    
    Parameters
    ----------
    left : bool
        left side of interval
    right : bool
        right side of interval
    Attributes
    ----------
    left : bool
        left side of interval
    right : bool
        right side of interval
        
    '''
    def __init__(self, left: bool ,right: bool = None):

        if right is None:
            right = left
            
        if left < right:
            self.left = left
            self.right = right
        else:
            self.left = left
            self.right = right

    def __bool__(self):

        if self.left == 0 and self.right == 0:
            return False
        else:
            return True

    def __repr__(self):

        if self.left == 0 and self.right == 0:
            return 'False'
        elif self.left == 1 and self.right == 1:
            return 'True'
        else:
            return "[%g, %g]"%(self.left,self.right)

    __str__ = __repr__


###
#   Defines copula functions. For plotting and use in convolutions
#
#   To Do:
#
#           -> Once completed, the Sigma, Tau and Rho convolutions may be defined
#
#
#       By: Ander Gray, University of Liverpool, ander.gray@liverpool.ac.uk
###

import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
from .interval import *
from scipy.stats import multivariate_normal as mvn
from scipy.stats import norm

__all__ = ['Copula','ClaGen','ClaInv','FGen','FInv','indep','perf','opp','Cla','F','Gau','pi','M','W','Frank','Clayton','Gaussian']
class Copula(object):

    def __init__(self, cdf=None, func=None, param = None):

        self.cdf = cdf          # Could also include density, however expensive to compute
        self.func = func        # If the functional form is known
        self.param = param      # parameter for func

    def __repr__(self):

        statement1 = "Arbitrary"
        statement2 = ""

        if self.func is not None:
            func = self.func
            if func == indep: func = "π"
            if func == perf: func = "M"
            if func == opp: func = "W"
            if func == Gau: func = "Gau"
            if func == Cla: func = "Clayton"
            if func == F: func = "Frank"
            statement1 = f'{func}'

        if self.param is not None:
            func = self.func
            parName = "par"
            if func == Gau: parName = "r"
            if func == F: parName = "s"
            if func == Cla: parName = "t"
            statement2 = f'{parName}={self.param}'

        return f'Copula ~ {statement1}({statement2})'

    def get_cdf(self, x, y):    # x, y are points on the unit square

        if self.func is not None:   # If function is known, return it
            if self.param is not None: return self.func(x, y, self.param)
            return self.func(x,y)

        else:   # Simple inner/outter interpolation. Returns interval for cdf value
            xsize, ysize = self.cdf.shape

            xIndexLower = int(np.floor(x * (xsize-1)))
            yIndexLower = int(np.floor(y * (ysize-1)))

            xIndexUpper = int(np.ceil(x * (xsize-1)))
            yIndexUpper = int(np.ceil(y * (ysize-1)))

            return Interval(self.cdf[xIndexLower, yIndexLower], self.cdf[xIndexUpper, yIndexUpper])

    def get_mass(self, x, y):   # x, y are intervals on the unit square

        C22 = self.get_cdf(x.hi(), y.hi())
        C21 = self.get_cdf(x.hi(), y.lo())
        C12 = self.get_cdf(x.lo(), y.hi())
        C11 = self.get_cdf(x.lo(), y.lo())

        return C22 - C21 - C12 + C11

    def show(self, pn = 50, fontsize = 20, cols = cm.RdGy):
        ##
        #   All the extra stuff is so that no more than 200 elements are plotted
        ##
        A = self.cdf; m = len(A)

        if m < pn: pn = m

        x = y = np.linspace(0, 1, num = pn)
        X, Y = np.meshgrid(x,y)

        nm = round(m/pn)
        Z = A[::nm,::nm]    # Skip over evelemts

        fig = plt.figure("SurfacPlots",figsize=(10,10))
        ax = fig.add_subplot(1,1,1,projection="3d")
        ax.plot_surface(X, Y, Z, rstride=2,edgecolors="k", cstride=2, alpha=0.8, linewidth=0.25, cmap = cols)
        plt.xlabel("X",fontsize = fontsize); plt.ylabel("Y", fontsize = fontsize)

        plt.show()

    def showContour(self, fontsize = 20, cols = cm.coolwarm):
        ##
        #   All the extra stuff is so that no more than 200 elements are plotted
        ##
        pn = 200    # Max plot number
        A = self.cdf; m = len(A)

        if m < pn: pn = m

        x = y = np.linspace(0, 1, num = pn)
        X, Y = np.meshgrid(x,y)

        nm = round(m/pn)
        Z = A[::nm,::nm]    # Skip over evelemts

        fig = plt.figure("SurfacPlots",figsize=(10,10))
        ax = fig.add_subplot(2,1,1,projection="3d")
        ax.plot_surface(X, Y, Z, rstride=2,edgecolors="k", cstride=2, alpha=0.8, linewidth=0.25, cmap = cols)
        plt.xlabel("X",fontsize = fontsize); plt.ylabel("Y", fontsize = fontsize)
        plt.title("Surface Plot", fontsize = fontsize)

        ax1 = fig.add_subplot(2,1,2)
        cp = ax1.contour(X, Y, Z, cmap = cols, levels = 15)
        ax1.clabel(cp, inline=1, fontsize=10)
        plt.xlabel("X", fontsize = fontsize); plt.ylabel("Y", fontsize = fontsize)
        plt.title("Contour Plot", fontsize = fontsize)

        plt.tight_layout()

        plt.show()

###
#   Copula generators for Archimedean (Frank and Clayton) copulas. Allows for easy and accurate copula generation
#   at any dimension in terms of univariate functions (generator and inverse generator).
###

def ClaGen(x, t = 1): return 1/t * (x **(-t) -1 )       # Clayton copula generator
def ClaInv(x, t = 1): return (1 + t * x) ** (-1/t)      # Inverse generator

def FGen(x, s = 1):                     # Frank generator
    X1 = np.exp( -(x * s) ) -1
    X2 = np.exp( -s ) -1
    return -np.log( X1 / X2)

def FInv(x, s = 1):                     # Inverse
    X1 = np.exp( -x ) * (np.exp(-s) - 1)
    X2 = np.log(1 + X1)
    return - (X2 / s)

###
#   Copula functions and constructors
###
def indep(x,y): return x*y
def perf(x,y): return min(x,y)
def opp(x,y): return max(x+y-1,0)
def Cla(x, y, t = 1): return ClaInv( ClaGen(x, t) + ClaGen(y, t), t)
def F(x,y,s = 1): return FInv( FGen(x, s) + FGen(y, s), s)
def Gau(x,y,r=0):
    if x == 0: return 0
    if y == 0: return 0
    return mvn.cdf([norm.ppf(x), norm.ppf(y)], mean = [0, 0], cov=[[1, r], [r, 1]])



# Copula constructors
def pi(steps = 200):
    x = y = np.linspace(0, 1, num=steps)
    cdf = np.array([[xs * ys for xs in x] for ys in y])
    return Copula(cdf, indep)

def M(steps = 200):
    x = y = np.linspace(0, 1, num=steps)
    cdf = np.array([[perf(xs, ys) for xs in x] for ys in y])
    return Copula(cdf, perf)

def W(steps = 200):
    x = y = np.linspace(0, 1, num=steps)
    cdf = np.array([[opp(xs, ys) for xs in x] for ys in y])
    return Copula(cdf, opp)

def Frank(s = 0, steps = 200):      #   s is real; inf for perfect, 0 for indep, -inf for oposite

    if s == float('-inf'):         # Limit should be set earlier
        C = W()
        return Copula(C.cdf, F, float('-inf'))
    if s == 0:
        C = pi()
        return Copula(C.cdf, F, 1)
    if s == float('inf'):
        C = M()
        return Copula(C.cdf, F, float('inf'))

    x = y = np.linspace(0, 1, num=steps)
    cdf = np.array([[F(xs, ys, s) for xs in x] for ys in y])
    return Copula(cdf, F, s)

def Clayton(t = 1, steps = 200):    #   t>-1; -1 for opposite, 0 for indep and inf for perfect

    if t == 0:
        C = pi()
        return Copula(C.cdf, Cla , 0)
    if t == -1:
        C = W()
        return Copula(C.cdf, Cla ,-1)
    if t == float('inf'):                       # Limit should be set earlier
        C = M()
        return Copula(C.cdf, Cla , float('inf'))

    x = np.linspace(0, 1, num=steps)
    xx, yy = np.meshgrid(x,x,indexing='ij')

    X = np.array([xx.flatten(), yy.flatten()])
    cdf = Cla(X[0, ], X[1, ], t)
    cdf = cdf.reshape(steps, steps)

    cdf[0,] = 0; cdf[:,0] = 0           # Grounds C

    return Copula(cdf, Cla, t)

def Gaussian(r = 0, steps = 200):   #   -1 <= r <=1 ; -1 for opposite, 1 for indep and 1 for perfect

    if r == 0:
        C = pi()
        return Copula(C.cdf, Gau, 0)
    if r == -1:
        C = W()
        return Copula(C.cdf, Gau, -1)
    if r == 1:
        C = M()
        return Copula(C.cdf, Gau, 1)

    x = np.linspace(0, 1, num=steps)
    xx, yy = np.meshgrid(x,x,indexing='ij')

    X = np.array([xx.flatten(), yy.flatten()])
    cdf = mvn.cdf(norm.ppf(X.transpose()), mean=[0,0], cov=[[1, r], [r, 1]])
    cdf = cdf.reshape(steps, steps)

    cdf[0,] = 0; cdf[:,0] = 0           # Grounds C

    return Copula(cdf, Gau, r)
